//	---------------------------------------------------------------------------
//	$Header: /fp/fp1/67NSYY1QMG4igynohtna/anthony/teestream.h,v 1.6 2002/06/01 15:25:32 ynohtna Exp $
//
//	(c) Copyright 2002 Anthony Bowyer-Lowe.  All rights reserved.
//
//	Anthony Bowyer-Lowe / ynohtna
//	ynohtna@ynohtna.org
//	http://www.ynohtna.org/
//	---------------------------------------------------------------------------
#if !defined(ALPHA_808_TEESTREAM_HEADER)
#define ALPHA_808_TEESTREAM_HEADER

#include <algorithm>
#include <ostream>
#include <sstream>
#include <string>
#include <vector>


//	---------------------------------------------------------------------------
//	A teestream is a stringstream that pipes its input onto zero or more
//	ostreams.
//	It's useful for relaying data from apis that only support a single ostream
//	onto multiple streams (simultaneously). Often found loitering in the
//	company of cout, cerr, logging files, and debug streams.
//
//	Usage:
//		anthony::teestream tee;
//		tee.attach(std::cerr);
//		tee.attach(logfile);
//		tee.attach(new debugstream);
//		tee << "Testing..." << std::endl;
//		tee << std::complex(0.0f, 2.3f) << std::endl;
//
//	Notes:
//	The streams used by the teestream need to have a lifetime at least as
//	long as the attached teestream.
//
//	A stream pointer passed to a teestream will become owned by the teestream,
//	and released at the teestream's time of death. This is to simplify
//	the aforementioned lifetime issues via simple syntax:
//		tee.attach(new consolestream);
//
//	The data within the teestream is only relayed onto the attached streams
//	when flushed (by inserting std::endl or std::flush, or by calling the
//	flush method of the teestream).
//
//	There is no capability at present to detach a stream from a teestream.
//
//	A teestream without attached streams will consume all input, and thus
//	function as a null_stream.
//	---------------------------------------------------------------------------
namespace anthony
{
#if 0
}	//	Bah fungoo to you, auto-indent.
#endif


//	Custom streambuf that propagates contents to associates upon sync ---------
template<typename E, typename T = std::char_traits<E> >
class basic_teebuf : public std::basic_stringbuf<E, T>
{
	typedef std::basic_ostream<E, T>	stream;

public:
	basic_teebuf()
	{	}
	~basic_teebuf() {
		sync();
	}
	void add(stream* s) {
		streams_.push_back(s);
	}
protected:
	typedef std::vector<stream*>		stream_list;
	stream_list streams_;
	struct outputter {
		const char* s_;
		outputter(const char* s) : s_(s) {	}
		void operator()(stream* os) {
			*os << s_;
			os->flush();
		}
	};
	int sync() {
		std::for_each(streams_.begin(), streams_.end(), outputter(str().c_str()));
		str(std::basic_string<E>());	//	Clear the string buffer.
		return 0;
	}
};


//	---------------------------------------------------------------------------
template<typename E, typename T = std::char_traits<E> >
class basic_teestream : public std::basic_ostream<E, T>
{
	typedef anthony::basic_teebuf<E, T> teebuf;
	typedef std::basic_ostream<E, T>	stream;
	typedef std::vector<stream*>		stream_ptrs;
	stream_ptrs							owned_streams_;

public:
	basic_teestream() : std::basic_ostream<E, T>(new teebuf())
	{	}
	~basic_teestream() {
		delete rdbuf();
		stream_ptrs::const_iterator itr = owned_streams_.begin();
		stream_ptrs::const_iterator end = owned_streams_.end();
		for(; itr != end; ++itr)
			delete *itr;				//	Destroy owned streams.
	}
	void attach(stream& s) {
		teebuf* tb = reinterpret_cast<teebuf*>(rdbuf());
		tb->add(&s);
	}
	void attach(stream* s) {			//	We claim ownership of this stream.
		teebuf* tb = reinterpret_cast<teebuf*>(rdbuf());
		tb->add(s);
		owned_streams_.push_back(s);	//	You belong to me now.
	}
};


//	---------------------------------------------------------------------------
typedef basic_teestream<char>		teestream;
typedef basic_teestream<wchar_t>	teestreamw;


//	---------------------------------------------------------------------------
}	//	namespace anthony


//	---------------------------------------------------------------------------
#endif	//	ALPHA_808_TEESTREAM_HEADER
